<?php

namespace Drupal\patternkit\Form;

use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Extension\ModuleHandlerInterface;
use Drupal\Core\Form\ConfigFormBase;
use Drupal\Core\Form\FormStateInterface;
use Drupal\editor\Entity\Editor;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Settings form for configuring the JSON Library Plugin.
 */
class PatternLibraryJSONForm extends ConfigFormBase {

  /**
   * The entity type manager service.
   *
   * @var \Drupal\Core\Entity\EntityTypeManagerInterface
   */
  protected EntityTypeManagerInterface $entityTypeManager;

  /**
   * The module handler service.
   *
   * @var \Drupal\Core\Extension\ModuleHandlerInterface
   */
  protected ModuleHandlerInterface $moduleHandler;

  /**
   * Settings identifier.
   *
   * @var string
   */
  public const SETTINGS = 'patternkit.settings';

  /**
   * Array of themes supported by the JSON Schema Editor.
   *
   * @var array
   *
   * @see https://github.com/json-editor/json-editor
   *
   * @todo Move to yml config.
   *
   * @todo Materialize doesn't support the Shadow dom: figure out encapsulation.
   */
  public const THEMES = [
    'barebones' => [],
    'html' => [],
    'jqueryui' => [
      'css' => 'https://code.jquery.com/ui/1.12.1/themes/base/jquery-ui.css',
    ],
    'bootstrap2' => [
      'css' => 'https://netdna.bootstrapcdn.com/twitter-bootstrap/2.3.2/css/bootstrap-combined.min.css',
    ],
    'bootstrap3' => [
      'css' => 'https://netdna.bootstrapcdn.com/bootstrap/3.3.7/css/bootstrap.min.css',
    ],
    'bootstrap4' => [
      'css' => 'library://patternkit/patternkit.jsoneditor.theme.bootstrap4/css',
    ],
    'cygnet' => [
      'css' => 'library://patternkit/cygnet/css',
      'css_no_shadow_dom' => 'library://patternkit/cygnet_no_shadow_dom/css',
    ],
    'foundation3' => [
      'css' => 'https://cdnjs.cloudflare.com/ajax/libs/foundation/3.2.5/stylesheets/foundation.css',
    ],
    'foundation4' => [
      'css' => 'https://cdnjs.cloudflare.com/ajax/libs/foundation/4.3.2/css/foundation.min.css',
    ],
    'foundation5' => [
      'css' => 'https://cdnjs.cloudflare.com/ajax/libs/foundation/5.5.3/css/foundation.min.css',
    ],
    'foundation6' => [
      'css' => 'https://cdnjs.cloudflare.com/ajax/libs/foundation/6.2.4/foundation.min.css',
    ],
  ];

  /**
   * Array of icons supported by the JSON Schema Editor.
   *
   * @var array
   *
   * @see https://github.com/json-editor/json-editor
   *
   * @todo Move to yml config.
   */
  public const ICONS = [
    'none' => '',
    'jqueryui' => '',
    'bootstrap2' => '',
    'bootstrap3' => '',
    'foundation2' => 'https://cdnjs.cloudflare.com/ajax/libs/foundicons/2.0/stylesheets/general_foundicons.css',
    'foundation3' => 'https://cdnjs.cloudflare.com/ajax/libs/foundicons/3.0.0/foundation-icons.css',
    'fontawesome3' => 'https://cdnjs.cloudflare.com/ajax/libs/font-awesome/3.2.1/css/font-awesome.css',
    'fontawesome4' => 'https://cdnjs.cloudflare.com/ajax/libs/font-awesome/4.0.3/css/font-awesome.css',
    'fontawesome5' => 'https://use.fontawesome.com/releases/v5.6.1/css/all.css',
  ];

  /**
   * Array of editor options supported by the JSON Schema Editor.
   *
   * @var array
   *
   * @see self::getAllowedWysiwygEditors()
   */
  public const EDITORS = [
    'ckeditor' => 'CKEditor',
    'ckeditor5' => 'CKEditor5',
    'prosemirror' => 'ProseMirror',
    'quill' => 'Quill',
  ];

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): self {
    $instance = parent::create($container);
    $instance->entityTypeManager = $container->get('entity_type.manager');
    $instance->moduleHandler = $container->get('module_handler');
    return $instance;
  }

  /**
   * {@inheritdoc}
   */
  public function buildForm(array $form, FormStateInterface $form_state) :array {
    $config = $this->config(static::SETTINGS);

    $themes = array_keys(static::THEMES);
    $form['patternkit_json_editor_theme'] = [
      '#type' => 'select',
      '#title' => $this->t('Select the JSON Editor Theme'),
      '#options' => array_combine($themes, $themes),
      '#default_value' => $config->get('patternkit_json_editor_theme'),
    ];

    $icons = array_keys(static::ICONS);
    $form['patternkit_json_editor_icons'] = [
      '#type' => 'select',
      '#title' => $this->t('Select the icons to be used with the editor'),
      '#options' => array_combine($icons, $icons),
      '#default_value' => $config->get('patternkit_json_editor_icons'),
    ];

    $form['patternkit_json_editor_css'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Add to the editor theme CSS'),
      '#description' => $this->t('Enter a comma-separated list of additional CSS to include.'),
      '#default_value' => $config->get('patternkit_json_editor_css'),
    ];

    $form['patternkit_json_editor_js'] = [
      '#type' => 'textfield',
      '#title' => $this->t('Add to the editor theme JS'),
      '#description' => $this->t('Enter a comma-separated list of additional JS to include.'),
      '#default_value' => $config->get('patternkit_json_editor_js'),
    ];

    $form['patternkit_json_editor_use_shadow_dom'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Use the Shadow DOM'),
      '#description' => $this->t('Select whether to use the Shadow DOM for the JSON
      Editor form. Most themes require the Shadow DOM, but most WYSIWYG editors
      do not work with it.'),
      '#default_value' => $config->get('patternkit_json_editor_use_shadow_dom') ?? TRUE,
    ];

    $form['patternkit_json_editor_disable_properties_buttons'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Hide the properties buttons'),
      '#description' => $this->t('Select whether to hide the property buttons (maps to json-editor option <em>disable_properties</em>).'),
      '#default_value' => $config->get('patternkit_json_editor_disable_properties_buttons') ?? FALSE,
    ];

    $form['patternkit_json_editor_disable_edit_json_button'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Hide the JSON Edit button'),
      '#description' => $this->t('Select whether to show the Edit JSON button (maps to json-editor option <em>disable_edit_json</em>).'),
      '#default_value' => $config->get('patternkit_json_editor_disable_edit_json_button') ?? TRUE,
    ];

    $form['patternkit_json_editor_cache_schemas'] = [
      '#type' => 'checkbox',
      '#title' => $this->t('Cache external schema references'),
      '#description' => $this->t('Select whether JSON Editor should cache schema lookups, which can speed up the editor form (maps to JSON Editor option <em>ajax_cache_responses</em>).'),
      '#default_value' => $config->get('patternkit_json_editor_cache_schemas') ?? FALSE,
    ];

    $form['patternkit_json_editor_wysiwyg'] = [
      '#type' => 'select',
      '#title' => $this->t('WYSIWYG editor'),
      '#options' => $this->getAllowedWysiwygEditors(),
      '#empty_value' => '',
      '#default_value' => $config->get('patternkit_json_editor_wysiwyg') ?: '',
    ];

    $ckeditor_toolbar_options = [];
    foreach ($this->entityTypeManager->getStorage('editor')->loadMultiple() as $editor) {
      $accepted_editors = [
        'ckeditor',
        'ckeditor5',
      ];
      /** @var \Drupal\editor\EditorInterface $editor */
      if (in_array($editor->getEditor(), $accepted_editors)) {
        $ckeditor_toolbar_options[$editor->id()] = $editor->label();
      }
    }
    asort($ckeditor_toolbar_options, SORT_NATURAL);

    $form['patternkit_json_editor_ckeditor_toolbar'] = [
      '#type' => 'select',
      '#title' => $this->t('CKEditor toolbar'),
      '#options' => $ckeditor_toolbar_options,
      '#empty_value' => '',
      '#default_value' => $config->get('patternkit_json_editor_ckeditor_toolbar') ?: '',
      '#states' => [
        'visible' => [
          ':input[name="patternkit_json_editor_wysiwyg"]' => [
            ['value' => 'ckeditor'],
            ['value' => 'ckeditor5'],
          ],
        ],
      ],
    ];

    return parent::buildForm($form, $form_state);
  }

  /**
   * {@inheritdoc}
   */
  protected function getEditableConfigNames(): array {
    return [static::SETTINGS];
  }

  /**
   * Get a list of options for allowed WYSIWYG editors to select.
   *
   * @return array
   *   An options list of WYSIWYG editors with all dependencies available.
   */
  protected function getAllowedWysiwygEditors(): array {
    // Remove options from supported editors if their required modules are not
    // enabled.
    $excluded = [];

    if (!$this->moduleHandler->moduleExists('ckeditor')) {
      $excluded[] = 'CKEditor';
    }
    if (!$this->moduleHandler->moduleExists('ckeditor5')) {
      $excluded[] = 'CKEditor5';
    }

    return array_diff(static::EDITORS, $excluded);
  }

  /**
   * {@inheritdoc}
   */
  public function getFormId(): string {
    return 'patternkit_json_editor_config';
  }

  /**
   * {@inheritdoc}
   */
  public function validateForm(array &$form, FormStateInterface $form_state) {
    // Validate editor and WYSIWYG toolbar selections.
    $wysiwyg = $form_state->getValue('patternkit_json_editor_wysiwyg');
    $toolbar = $form_state->getValue('patternkit_json_editor_ckeditor_toolbar');

    if ($toolbar && in_array($wysiwyg, ['ckeditor', 'ckeditor5'])) {
      $editor = Editor::load($toolbar);

      if ($editor->getEditor() !== $wysiwyg) {
        $form_state->setErrorByName('patternkit_json_editor_ckeditor_toolbar', $this->t(
          'The selected toolbar is not configured to use the selected toolbar. Expected to find "@expected", but found "@actual" instead.',
          [
            '@expected' => $wysiwyg,
            '@actual' => $editor->getEditor(),
          ]
        ));
      }
    }
  }

  /**
   * {@inheritdoc}
   */
  public function submitForm(array &$form, FormStateInterface $form_state): void {
    $form_values = $form_state->getValues();
    $config = $this->config(self::SETTINGS);
    $config->set('patternkit_json_editor_theme', $form_values['patternkit_json_editor_theme']);
    $config->set('patternkit_json_editor_icons', $form_values['patternkit_json_editor_icons']);
    $config->set('patternkit_json_editor_css', $form_values['patternkit_json_editor_css']);
    $config->set('patternkit_json_editor_js', $form_values['patternkit_json_editor_js']);
    $config->set('patternkit_json_editor_use_shadow_dom', (bool) $form_values['patternkit_json_editor_use_shadow_dom']);
    $config->set('patternkit_json_editor_disable_properties_buttons', (bool) $form_values['patternkit_json_editor_disable_properties_buttons']);
    $config->set('patternkit_json_editor_disable_edit_json_button', (bool) $form_values['patternkit_json_editor_disable_edit_json_button']);
    $config->set('patternkit_json_editor_cache_schemas', (bool) $form_values['patternkit_json_editor_cache_schemas']);
    $config->set('patternkit_json_editor_wysiwyg', $form_values['patternkit_json_editor_wysiwyg']);
    $config->set('patternkit_json_editor_ckeditor_toolbar', $form_values['patternkit_json_editor_ckeditor_toolbar']);
    $config->save();

    parent::submitForm($form, $form_state);
  }

}
